'From Objectworks for Smalltalk-80(tm), Version 2.5 of 29 July 1989 on 22 March 1995 at 8:48:28 pm'!

SequenceableCollection subclass: #Interval
	instanceVariableNames: 'start stop step '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Sequenceable'!
Interval comment:
'The class Interval represents a finite arithmetic progression.

Because of the inevitable round-off errors which occur with floating-point arithmetic,
intervals which have floating-point values may have surprising properties.  For example:
	| big |
	big := 1.0e7.
	(big to: big + 1 by: 0.25) collect: [:v | v - big]
...yields (0.0 0.0 0.0 1.0 1.0) rather than (0.0 0.25 0.5 0.75 1.0) as you''d expect.
This is because there is an unavoidable roundoff error when computing concrete
elements of the interval since --
	1.0e7 + (2 * 0.25) = 1.0e7
but --
	1.0e7 + (3 * 0.25) = (1.0e7 + 1)
...hence the roundoff error is distributed unevenly (although monotonically) over the
elements generated by the collect:. 

Instance Variables:
	start	<Number>	Initial number in the progression
	stop	<Number>	Last number in the progression
	step	<Number>	Increment for determining the next number in the progression'!


!Interval methodsFor: 'comparing'!

= anInterval 
	"Answer true if anInterval is kind of my class, and if our  
	starts, steps and sizes are equal.  
	If anInterval is not kind of my class, superclass will deside  
	we are = or not."

	(anInterval isKindOf: self class)
		ifTrue: [^start = anInterval first
					and: [step = anInterval increment
					and: [self size = anInterval size]]]
		ifFalse: [^super = anInterval]!

hash
	"Answer a SmallInteger unique to the receiver."

	^(((start hash bitShift: 2)
		bitOr: stop hash)
		bitShift: 1)
		bitOr: self size! !

!Interval methodsFor: 'accessing'!

at: anInteger 
	"Answer the anInteger'th element."

	(anInteger >= 1 and: [anInteger <= self size])
		ifTrue: [^start + (step * (anInteger - 1))]
		ifFalse: [^self subscriptBoundsError: anInteger]!

at: anInteger put: anObject 
	"Provide an error notification that storing into an Interval is not 
	allowed. "

	self error: 'you can not store into an interval'!

first
	"Answer the first element of the arithmetic progression."

	^start!

increment
	"Answer the receiver's interval increment."

	^step!

last
	"Answer the last element of the arithmetic progression."

	^stop - (stop - start \\ step)!

size
	"Answer how many elements the receiver contains."
	"E.g. what is the distance between start and stop measured
	in step units."

	^step < 0
		ifTrue: [start < stop
				ifTrue: [0]
				ifFalse: [stop - start // step + 1]]
		ifFalse: [stop < start
				ifTrue: [0]
				ifFalse: [stop - start // step + 1]]! !

!Interval methodsFor: 'adding'!

add: newObject 
	"Provide an error notification that adding to an Interval is not allowed."

	self shouldNotImplement! !

!Interval methodsFor: 'removing'!

remove: newObject 
	"Provide an error notification that removing an element
	 from an Interval is not allowed."

	self error: 'elements cannot be removed from an Interval'! !

!Interval methodsFor: 'copying'!

copy
	"Answer another instance just like the receiver."
	"Intervals override this message because their species is Array and 
	copy, as inherited from SequenceableCollection, uses copyFrom:to:, 
	which creates a new object of the species."

	^self shallowCopy! !

!Interval methodsFor: 'enumerating'!

collect: aBlock 
	"Evaluate aBlock with each of the values of the receiver as the 
	argument. Collect the resulting values into a collection that is like 
	the receiver. Answer the new collection."
	"Override implementation in SequencableCollection to avoid self>>at:. 
	We avoid accumulating a delta to avoid round-off errors."

	| result i |
	result _ self species new: self size.
	i := 0.
	self do: [:v | result at: (i := i + 1) put: (aBlock value: v)].
	^result!

do: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument."
	"We avoid accumulating a delta to minimize round-off error."

	0 to: self size - 1 do: [:n | aBlock value: start + (step * n)]!

reverseDo: aBlock 
	"Evaluate aBlock for each element of the receiver's interval, in reverse order."
	"We avoid accumulating a delta to minimize round-off error."

	self size-1 to: 0 by: -1 do: [:n | aBlock value: start + (step * n)]! !

!Interval methodsFor: 'printing'!

printOn: aStream 
	"Append to the argument aStream a sequence of characters 
	that identifies the interval.  The literal format for Intervals 
	is used, that is
		(startNumber to: stopNumber by: stepNumber)
	unless step is 1 in which case the format is
		(startNumber to: stopNumber)"

	aStream nextPut: $(.
	start printOn: aStream.
	aStream nextPutAll: ' to: '.
	stop printOn: aStream.
	step ~= 1
		ifTrue: 
			[aStream nextPutAll: ' by: '.
			step printOn: aStream].
	aStream nextPut: $)!

storeOn: aStream 
	"Because numbers store and print the same, append to
	the argument aStream the literal description of an Interval,
	same as printOn:."

	self printOn: aStream! !

!Interval methodsFor: 'private: private'!

setFrom: startInteger to: stopInteger by: stepInteger 
	"Initialize the instance variables."

	start _ startInteger.
	stop _ stopInteger.
	step _ stepInteger!

species
	"Answer the preferred class for reconstructing the receiver,
	that is, Array."

	^Array! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Interval class
	instanceVariableNames: ''!


!Interval class methodsFor: 'instance creation'!

from: startInteger to: stopInteger 
	"Answer a new instance of Interval, starting at startInteger, ending and
	stopInteger, and with an interval increment of 1."

	^self basicNew
		setFrom: startInteger
		to: stopInteger
		by: 1!

from: startInteger to: stopInteger by: stepInteger 
	"Answer a new instance of Interval, starting at startInteger, ending and
	stopInteger, and with an interval increment of stepInteger."

	^self basicNew
		setFrom: startInteger
		to: stopInteger
		by: stepInteger!

new
	"Provide an error notification that Intervals are not created
	using this message."

	self error: 'Intervals are created with from:to: and from:to:by:'! !